@@ -69,6 +69,34 @@ impl Gate {
     /// Mask to get the state value from `u8`.
     const STATE_MASK: u8 = 0b11;
 
+    /// Creates a new [`Gate`].
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use saa::Gate;
+    ///
+    /// let gate = Gate::new();
+    /// ```
+    #[cfg(not(feature = "loom"))]
+    #[inline]
+    #[must_use]
+    pub const fn new() -> Self {
+        Self {
+            state: AtomicUsize::new(0),
+        }
+    }
+
+    /// Creates a new [`Gate`].
+    #[cfg(feature = "loom")]
+    #[inline]
+    #[must_use]
+    pub fn new() -> Self {
+        Self {
+            state: AtomicUsize::new(0),
+        }
+    }
+
     /// Returns the current state of the [`Gate`].
     ///
     /// # Examples
@@ -299,41 +327,6 @@ impl Gate {
         pinned_pager.poll_async().await
     }
 
-    /// Enters the [`Gate`] asynchronously with a wait callback.
-    ///
-    /// Returns the current state of the [`Gate`].
-    ///
-    /// # Errors
-    ///
-    /// Returns an [`Error`] if it failed to enter the [`Gate`].
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use futures::future;
-    /// use saa::Gate;
-    ///
-    /// let gate = Gate::default();
-    ///
-    /// let a = async {
-    ///     let mut wait = false;
-    ///     assert!(gate.enter_async_with(|| wait = true).await.is_ok());
-    /// };
-    /// let b = async {
-    ///    gate.permit();
-    /// };
-    /// future::join(a, b);
-    /// ```
-    #[inline]
-    pub async fn enter_async_with<F: FnOnce()>(&self, wait_callback: F) -> Result<State, Error> {
-        let mut pinned_pager = pin!(Pager::default());
-        pinned_pager
-            .wait_queue()
-            .construct(self, Opcode::Wait(0), false);
-        self.push_wait_queue_entry(&mut pinned_pager, wait_callback);
-        pinned_pager.poll_async().await
-    }
-
     /// Enters the [`Gate`] synchronously.
     ///
     /// Returns the current state of the [`Gate`].
@@ -363,10 +356,10 @@ impl Gate {
     ///     assert_eq!(gate_clone.enter_sync(), Ok(State::Controlled));
     /// });
     ///
-    /// let mut cnt = 0;
-    /// while cnt != 2 {
+    /// let mut count = 0;
+    /// while count != 2 {
     ///     if let Ok(n) = gate.permit() {
-    ///         cnt += n;
+    ///         count += n;
     ///     }
     /// }
     ///
@@ -378,10 +371,46 @@ impl Gate {
         self.enter_sync_with(|| ())
     }
 
+    /// Enters the [`Gate`] asynchronously with a wait callback.
+    ///
+    /// Returns the current state of the [`Gate`]. The callback is invoked when the task starts
+    /// waiting.
+    ///
+    /// # Errors
+    ///
+    /// Returns an [`Error`] if it failed to enter the [`Gate`].
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use futures::future;
+    /// use saa::Gate;
+    ///
+    /// let gate = Gate::default();
+    ///
+    /// let a = async {
+    ///     let mut wait = false;
+    ///     assert!(gate.enter_async_with(|| wait = true).await.is_ok());
+    /// };
+    /// let b = async {
+    ///    gate.permit();
+    /// };
+    /// future::join(a, b);
+    /// ```
+    #[inline]
+    pub async fn enter_async_with<F: FnOnce()>(&self, begin_wait: F) -> Result<State, Error> {
+        let mut pinned_pager = pin!(Pager::default());
+        pinned_pager
+            .wait_queue()
+            .construct(self, Opcode::Wait(0), false);
+        self.push_wait_queue_entry(&mut pinned_pager, begin_wait);
+        pinned_pager.poll_async().await
+    }
+
     /// Enters the [`Gate`] synchronously with a wait callback.
     ///
-    /// Returns the current state of the [`Gate`].
-    ///
+    /// Returns the current state of the [`Gate`]. The callback is invoked when the task starts
+    /// waiting.
     /// # Errors
     ///
     /// Returns an [`Error`] if it failed to enter the [`Gate`].
@@ -409,10 +438,10 @@ impl Gate {
     ///     assert_eq!(gate_clone.enter_sync_with(|| wait = true), Ok(State::Controlled));
     /// });
     ///
-    /// let mut cnt = 0;
-    /// while cnt != 2 {
+    /// let mut count = 0;
+    /// while count != 2 {
     ///     if let Ok(n) = gate.permit() {
-    ///         cnt += n;
+    ///         count += n;
     ///     }
     /// }
     ///
@@ -420,12 +449,12 @@ impl Gate {
     /// thread_2.join().unwrap();
     /// ```
     #[inline]
-    pub fn enter_sync_with<F: FnOnce()>(&self, wait_callback: F) -> Result<State, Error> {
+    pub fn enter_sync_with<F: FnOnce()>(&self, begin_wait: F) -> Result<State, Error> {
         let mut pinned_pager = pin!(Pager::default());
         pinned_pager
             .wait_queue()
             .construct(self, Opcode::Wait(0), true);
-        self.push_wait_queue_entry(&mut pinned_pager, wait_callback);
+        self.push_wait_queue_entry(&mut pinned_pager, begin_wait);
         pinned_pager.poll_sync()
     }
 
@@ -509,21 +538,15 @@ impl Gate {
 
     /// Pushes the wait queue entry.
     #[inline]
-    fn push_wait_queue_entry<F: FnOnce()>(
-        &self,
-        pager: &mut Pin<&mut Pager<Self>>,
-        mut wait_callback: F,
-    ) {
+    fn push_wait_queue_entry<F: FnOnce()>(&self, pager: &mut Pin<&mut Pager<Self>>, begin_wait: F) {
         loop {
             let state = self.state.load(Acquire);
             match State::from(state & WaitQueue::DATA_MASK) {
                 State::Controlled => {
-                    if let Some(returned) =
-                        self.try_push_wait_queue_entry(pager.wait_queue(), state, wait_callback)
-                    {
-                        wait_callback = returned;
+                    if !self.try_push_wait_queue_entry(pager.wait_queue(), state) {
                         continue;
                     }
+                    begin_wait();
                 }
                 State::Sealed => {
                     pager
