@@ -1,22 +1,23 @@
 //! [`Lock`] is a low-level locking primitive for both synchronous and asynchronous operations.
 
-#![deny(unsafe_code)]
-
 use std::fmt;
-use std::pin::{Pin, pin};
+use std::pin::Pin;
 #[cfg(not(feature = "loom"))]
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering::{self, AcqRel, Acquire, Relaxed, Release};
+#[cfg(not(feature = "loom"))]
 use std::thread::yield_now;
 
 #[cfg(feature = "loom")]
 use loom::sync::atomic::AtomicUsize;
+#[cfg(feature = "loom")]
+use loom::thread::yield_now;
 
 use crate::Pager;
 use crate::opcode::Opcode;
 use crate::pager::{self, SyncResult};
 use crate::sync_primitive::SyncPrimitive;
-use crate::wait_queue::{Entry, PinnedEntry, WaitQueue};
+use crate::wait_queue::{Entry, WaitQueue};
 
 /// [`Lock`] is a low-level locking primitive for both synchronous and asynchronous operations.
 ///
@@ -39,7 +40,7 @@ pub enum Mode {
     /// Waits for the [`Lock`] to be free or poisoned.
     ///
     /// [`Self::WaitExclusive`], [`Self::WaitShared`], [`Lock::try_lock`], and [`Lock::try_share`]
-    /// provide a way to bypass the fair queueing mechanism without spinning.
+    /// provide a way to bypass the fair queuing mechanism without spinning.
     WaitExclusive,
     /// Waits for a shared lock to be available or the [`Lock`] to be poisoned.
     ///
@@ -64,6 +65,37 @@ impl Lock {
     /// Poisoned error code.
     const POISONED: u8 = 2_u8;
 
+    /// Maximum spin count before waiting.
+    const MAX_SPIN_COUNT: usize = 128;
+
+    /// Creates a new [`Lock`].
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use saa::Lock;
+    ///
+    /// let lock = Lock::new();
+    /// ```
+    #[cfg(not(feature = "loom"))]
+    #[inline]
+    #[must_use]
+    pub const fn new() -> Self {
+        Self {
+            state: AtomicUsize::new(0),
+        }
+    }
+
+    /// Creates a new [`Lock`].
+    #[cfg(feature = "loom")]
+    #[inline]
+    #[must_use]
+    pub fn new() -> Self {
+        Self {
+            state: AtomicUsize::new(0),
+        }
+    }
+
     /// Returns `true` if the lock is currently free.
     ///
     /// # Examples
@@ -166,32 +198,7 @@ impl Lock {
     /// ```
     #[inline]
     pub async fn lock_async(&self) -> bool {
-        self.acquire_async_with_internal::<_, true>(|| {}).await
-    }
-
-    /// Acquires an exclusive lock asynchronously with a wait callback.
-    ///
-    /// Returns `false` if the lock is poisoned. The callback is invoked when the task starts
-    /// waiting for a lock.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use saa::Lock;
-    /// use std::sync::atomic::Ordering::Relaxed;
-    ///
-    /// let lock = Lock::default();
-    ///
-    /// async {
-    ///     lock.lock_async().await;
-    ///     assert!(lock.is_locked(Relaxed));
-    ///     assert!(!lock.is_shared(Relaxed))
-    /// };
-    /// ```
-    #[inline]
-    pub async fn lock_async_with<F: FnOnce()>(&self, begin_wait: F) -> bool {
-        self.acquire_async_with_internal::<_, true>(begin_wait)
-            .await
+        self.lock_async_with(|| ()).await
     }
 
     /// Acquires an exclusive lock synchronously.
@@ -216,6 +223,48 @@ impl Lock {
         self.lock_sync_with(|| ())
     }
 
+    /// Acquires an exclusive lock asynchronously with a wait callback.
+    ///
+    /// Returns `false` if the lock is poisoned. The callback is invoked when the task starts
+    /// waiting for a lock.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use saa::Lock;
+    /// use std::sync::atomic::Ordering::Relaxed;
+    ///
+    /// let lock = Lock::default();
+    ///
+    /// async {
+    ///     lock.lock_async().await;
+    ///     assert!(lock.is_locked(Relaxed));
+    ///     assert!(!lock.is_shared(Relaxed))
+    /// };
+    /// ```
+    #[inline]
+    pub async fn lock_async_with<F: FnOnce()>(&self, begin_wait: F) -> bool {
+        loop {
+            let (mut result, state) = self.try_lock_internal_spin();
+            if result == Self::ACQUIRED {
+                return true;
+            } else if result == Self::POISONED {
+                return false;
+            }
+            debug_assert_eq!(result, Self::NOT_ACQUIRED);
+
+            let async_wait = WaitQueue::default();
+            let async_wait_pinned = async_wait.pin();
+            async_wait_pinned.construct(self, Opcode::Exclusive, false);
+            if self.try_push_wait_queue_entry(async_wait_pinned, state) {
+                begin_wait();
+                result = async_wait_pinned.await;
+                debug_assert!(result == Self::ACQUIRED || result == Self::POISONED);
+                return result == Self::ACQUIRED;
+            }
+        }
+    }
+
     /// Acquires an exclusive lock synchronously with a wait callback.
     ///
     /// Returns `false` if the lock is poisoned. The callback is invoked when the task starts
@@ -237,7 +286,7 @@ impl Lock {
     #[inline]
     pub fn lock_sync_with<F: FnOnce()>(&self, mut begin_wait: F) -> bool {
         loop {
-            let (result, state) = self.try_lock_internal();
+            let (result, state) = self.try_lock_internal_spin();
             if result == Self::ACQUIRED {
                 return true;
             } else if result == Self::POISONED {
@@ -272,7 +321,7 @@ impl Lock {
     /// ```
     #[inline]
     pub fn try_lock(&self) -> bool {
-        self.try_lock_internal().0 == Self::ACQUIRED
+        self.try_lock_internal(self.state.load(Acquire)).0 == Self::ACQUIRED
     }
 
     /// Acquires a shared lock asynchronously.
@@ -295,32 +344,7 @@ impl Lock {
     /// ```
     #[inline]
     pub async fn share_async(&self) -> bool {
-        self.acquire_async_with_internal::<_, false>(|| ()).await
-    }
-
-    /// Acquires a shared lock asynchronously with a wait callback.
-    ///
-    /// Returns `false` if the lock is poisoned. The callback is invoked when the task starts
-    /// waiting for a lock.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use saa::Lock;
-    /// use std::sync::atomic::Ordering::Relaxed;
-    ///
-    /// let lock = Lock::default();
-    ///
-    /// async {
-    ///     lock.share_async().await;
-    ///     assert!(!lock.is_locked(Relaxed));
-    ///     assert!(lock.is_shared(Relaxed))
-    /// };
-    /// ```
-    #[inline]
-    pub async fn share_async_with<F: FnOnce()>(&self, begin_wait: F) -> bool {
-        self.acquire_async_with_internal::<_, false>(begin_wait)
-            .await
+        self.share_async_with(|| ()).await
     }
 
     /// Acquires a shared lock synchronously.
@@ -345,6 +369,48 @@ impl Lock {
         self.share_sync_with(|| ())
     }
 
+    /// Acquires a shared lock asynchronously with a wait callback.
+    ///
+    /// Returns `false` if the lock is poisoned. The callback is invoked when the task starts
+    /// waiting for a lock.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use saa::Lock;
+    /// use std::sync::atomic::Ordering::Relaxed;
+    ///
+    /// let lock = Lock::default();
+    ///
+    /// async {
+    ///     lock.share_async().await;
+    ///     assert!(!lock.is_locked(Relaxed));
+    ///     assert!(lock.is_shared(Relaxed))
+    /// };
+    /// ```
+    #[inline]
+    pub async fn share_async_with<F: FnOnce()>(&self, begin_wait: F) -> bool {
+        loop {
+            let (mut result, state) = self.try_share_internal_spin();
+            if result == Self::ACQUIRED {
+                return true;
+            } else if result == Self::POISONED {
+                return false;
+            }
+            debug_assert_eq!(result, Self::NOT_ACQUIRED);
+
+            let async_wait = WaitQueue::default();
+            let async_wait_pinned = async_wait.pin();
+            async_wait_pinned.construct(self, Opcode::Shared, false);
+            if self.try_push_wait_queue_entry(async_wait_pinned, state) {
+                begin_wait();
+                result = async_wait_pinned.await;
+                debug_assert!(result == Self::ACQUIRED || result == Self::POISONED);
+                return result == Self::ACQUIRED;
+            }
+        }
+    }
+
     /// Acquires a shared lock synchronously with a wait callback.
     ///
     /// Returns `false` if the lock is poisoned. The callback is invoked when the task starts
@@ -366,7 +432,7 @@ impl Lock {
     #[inline]
     pub fn share_sync_with<F: FnOnce()>(&self, mut begin_wait: F) -> bool {
         loop {
-            let (result, state) = self.try_share_internal();
+            let (result, state) = self.try_share_internal_spin();
             if result == Self::ACQUIRED {
                 return true;
             } else if result == Self::POISONED {
@@ -402,7 +468,7 @@ impl Lock {
     /// ```
     #[inline]
     pub fn try_share(&self) -> bool {
-        self.try_share_internal().0 == Self::ACQUIRED
+        self.try_share_internal(self.state.load(Acquire)).0 == Self::ACQUIRED
     }
 
     /// Registers a [`Pager`] to allow it to get an exclusive lock or a shared lock remotely.
@@ -450,8 +516,8 @@ impl Lock {
 
         loop {
             let (result, state) = match mode {
-                Mode::Exclusive => self.try_lock_internal(),
-                Mode::Shared => self.try_share_internal(),
+                Mode::Exclusive => self.try_lock_internal_spin(),
+                Mode::Shared => self.try_share_internal_spin(),
                 Mode::WaitExclusive | Mode::WaitShared => {
                     let state = self.state.load(Acquire);
                     let result = if state == usize::from(Self::POISONED) {
@@ -473,10 +539,7 @@ impl Lock {
                 break;
             }
 
-            if self
-                .try_push_wait_queue_entry(pager.wait_queue(), state, || ())
-                .is_none()
-            {
+            if self.try_push_wait_queue_entry(pager.wait_queue(), state) {
                 break;
             }
         }
@@ -615,109 +678,93 @@ impl Lock {
     /// ```
     #[inline]
     pub fn release_share(&self) -> bool {
-        match self.state.compare_exchange(1, 0, Release, Relaxed) {
+        match self.state.fetch_update(Release, Relaxed, |state| {
+            if state != 0 && state <= Self::MAX_SHARED_OWNERS {
+                Some(state - 1)
+            } else {
+                None
+            }
+        }) {
             Ok(_) => true,
             Err(state) => self.release_loop(state, Opcode::Shared),
         }
     }
 
-    /// Tries to acquire an exclusive lock.
+    /// Tries to acquire an exclusive lock with spin-backoff.
     #[inline]
-    fn try_lock_internal(&self) -> (u8, usize) {
-        let Err(state) = self
+    fn try_lock_internal_spin(&self) -> (u8, usize) {
+        let Err(mut state) = self
             .state
             .compare_exchange(0, WaitQueue::DATA_MASK, Acquire, Acquire)
         else {
             return (Self::ACQUIRED, 0);
         };
-        self.try_lock_internal_slow(state)
+        let mut result = Self::NOT_ACQUIRED;
+        for spin_count in 0..Self::MAX_SPIN_COUNT {
+            (result, state) = self.try_lock_internal(state);
+            if result != Self::NOT_ACQUIRED {
+                return (result, state);
+            }
+            if spin_count % 4 == 0 {
+                yield_now();
+            }
+        }
+        (result, state)
     }
 
-    /// Tries to acquire an exclusive lock, slowly.
-    fn try_lock_internal_slow(&self, mut state: usize) -> (u8, usize) {
+    /// Tries to acquire an exclusive lock.
+    #[inline]
+    fn try_lock_internal(&self, mut state: usize) -> (u8, usize) {
         loop {
             if state == Self::POISONED_STATE {
                 return (Self::POISONED, state);
-            } else if state & WaitQueue::ADDR_MASK != 0 || state & WaitQueue::DATA_MASK != 0 {
-                // There is a waiting thread, so this thread should not acquire the lock.
+            } else if state != 0 {
                 return (Self::NOT_ACQUIRED, state);
             }
-
-            if state & WaitQueue::DATA_MASK == 0 {
-                match self.state.compare_exchange(
-                    state,
-                    state | WaitQueue::DATA_MASK,
-                    Acquire,
-                    Acquire,
-                ) {
-                    Ok(_) => return (Self::ACQUIRED, 0),
-                    Err(new_state) => state = new_state,
-                }
+            match self
+                .state
+                .compare_exchange(0, WaitQueue::DATA_MASK, Acquire, Acquire)
+            {
+                Ok(_) => return (Self::ACQUIRED, 0),
+                Err(new_state) => state = new_state,
             }
         }
     }
 
-    /// Acquires an exclusive lock or a shared lock asynchronously.
+    /// Tries to acquire a shared lock with spin-backoff.
     #[inline]
-    async fn acquire_async_with_internal<F: FnOnce(), const EXCLUSIVE: bool>(
-        &self,
-        mut begin_wait: F,
-    ) -> bool {
-        loop {
-            let (mut result, state) = if EXCLUSIVE {
-                self.try_lock_internal()
+    fn try_share_internal_spin(&self) -> (u8, usize) {
+        let Err(mut state) = self.state.fetch_update(Acquire, Acquire, |state| {
+            if state < Self::MAX_SHARED_OWNERS {
+                Some(state + 1)
             } else {
-                self.try_share_internal()
-            };
-            if result == Self::ACQUIRED {
-                return true;
-            } else if result == Self::POISONED {
-                return false;
+                None
             }
-            debug_assert_eq!(result, Self::NOT_ACQUIRED);
-            debug_assert!(state & WaitQueue::ADDR_MASK != 0 || state & WaitQueue::DATA_MASK != 0);
-
-            let mode = if EXCLUSIVE {
-                Opcode::Exclusive
-            } else {
-                Opcode::Shared
-            };
-            let async_wait = pin!(WaitQueue::default());
-            async_wait.as_ref().construct(self, mode, false);
-            if let Some(returned) =
-                self.try_push_wait_queue_entry(async_wait.as_ref(), state, begin_wait)
-            {
-                begin_wait = returned;
-                continue;
+        }) else {
+            return (Self::ACQUIRED, 0);
+        };
+        let mut result = Self::NOT_ACQUIRED;
+        for spin_count in 0..Self::MAX_SPIN_COUNT {
+            (result, state) = self.try_share_internal(state);
+            if result != Self::NOT_ACQUIRED {
+                return (result, state);
+            }
+            if spin_count % 4 == 0 {
+                yield_now();
             }
-
-            result = PinnedEntry(Pin::new(async_wait.entry())).await;
-            debug_assert!(result == Self::ACQUIRED || result == Self::POISONED);
-            return result == Self::ACQUIRED;
         }
+        (result, state)
     }
 
     /// Tries to acquire a shared lock.
     #[inline]
-    fn try_share_internal(&self) -> (u8, usize) {
-        let Err(state) = self.state.compare_exchange(0, 1, Acquire, Acquire) else {
-            return (Self::ACQUIRED, 0);
-        };
-        self.try_share_internal_slow(state)
-    }
-
-    /// Tries to acquire a shared lock, slowly.
-    fn try_share_internal_slow(&self, mut state: usize) -> (u8, usize) {
+    fn try_share_internal(&self, mut state: usize) -> (u8, usize) {
         loop {
             if state == Self::POISONED_STATE {
                 return (Self::POISONED, state);
-            } else if state & WaitQueue::ADDR_MASK != 0
-                || state & WaitQueue::DATA_MASK >= Self::MAX_SHARED_OWNERS
-            {
-                // There is a waiting thread, or the lock can no longer be shared.
+            } else if state >= Self::MAX_SHARED_OWNERS {
                 return (Self::NOT_ACQUIRED, state);
             }
-
             match self
                 .state
                 .compare_exchange(state, state + 1, Acquire, Acquire)
