@@ -6,7 +6,7 @@ use std::cell::UnsafeCell;
 use std::marker::{PhantomData, PhantomPinned};
 use std::pin::Pin;
 
-use crate::wait_queue::{Entry, PinnedEntry, WaitQueue};
+use crate::wait_queue::{Entry, WaitQueue};
 
 /// Tasks holding a [`Pager`] can remotely acquire a desired resource.
 ///
@@ -97,9 +97,7 @@ impl<'s, S: SyncResult> Pager<'s, S> {
         if !self.is_registered() {
             return S::to_result(0, Some(Error::NotRegistered));
         }
-        let wait_queue = self.wait_queue();
-        let pinned_entry = PinnedEntry(Pin::new(wait_queue.entry()));
-        let result = pinned_entry.await;
+        let result = self.wait_queue().await;
         if result == Entry::ERROR_WRONG_MODE {
             return S::to_result(result, Some(Error::WrongMode));
         }
@@ -176,6 +174,6 @@ impl<'s, S: SyncResult> Pager<'s, S> {
     /// Returns a reference to the wait queue entry.
     #[inline]
     pub(crate) fn wait_queue(&self) -> Pin<&WaitQueue> {
-        WaitQueue::pinned_wait_queue(self.wait_queue.get())
+        WaitQueue::pin_ptr(self.wait_queue.get())
     }
 }
