@@ -4,7 +4,7 @@
 #![deny(unsafe_code)]
 
 use std::fmt;
-use std::pin::{Pin, pin};
+use std::pin::Pin;
 #[cfg(not(feature = "loom"))]
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering::{self, Acquire, Relaxed, Release};
@@ -16,7 +16,7 @@ use crate::Pager;
 use crate::opcode::Opcode;
 use crate::pager::{self, SyncResult};
 use crate::sync_primitive::SyncPrimitive;
-use crate::wait_queue::{Entry, PinnedEntry, WaitQueue};
+use crate::wait_queue::{Entry, WaitQueue};
 
 /// [`Semaphore`] is a synchronization primitive that allows a fixed number of threads to access a
 /// resource concurrently.
@@ -30,6 +30,34 @@ impl Semaphore {
     /// Maximum number of concurrent owners.
     pub const MAX_PERMITS: usize = WaitQueue::DATA_MASK;
 
+    /// Creates a new [`Semaphore`].
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use saa::Semaphore;
+    ///
+    /// let semaphore = Semaphore::new();
+    /// ```
+    #[cfg(not(feature = "loom"))]
+    #[inline]
+    #[must_use]
+    pub const fn new() -> Self {
+        Self {
+            state: AtomicUsize::new(0),
+        }
+    }
+
+    /// Creates a new [`Semaphore`].
+    #[cfg(feature = "loom")]
+    #[inline]
+    #[must_use]
+    pub fn new() -> Self {
+        Self {
+            state: AtomicUsize::new(0),
+        }
+    }
+
     /// Creates a new [`Semaphore`] with the given number of initially available permits.
     ///
     /// The maximum number of available permits is [`MAX_PERMITS`](Self::MAX_PERMITS), and if a
@@ -138,7 +166,25 @@ impl Semaphore {
     /// ```
     #[inline]
     pub async fn acquire_async(&self) {
-        self.acquire_async_with_internal(1, || {}).await;
+        self.acquire_many_async_with(1, || {}).await;
+    }
+
+    /// Gets a permit from the semaphore synchronously.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use saa::Semaphore;
+    /// use std::sync::atomic::Ordering::Relaxed;
+    ///
+    /// let semaphore = Semaphore::default();
+    ///
+    /// semaphore.acquire_sync();
+    /// assert_eq!(semaphore.available_permits(Relaxed), Semaphore::MAX_PERMITS - 1);
+    /// ```
+    #[inline]
+    pub fn acquire_sync(&self) {
+        self.acquire_many_sync_with(1, || ());
     }
 
     /// Gets a permit from the semaphore asynchronously with a wait callback.
@@ -162,25 +208,7 @@ impl Semaphore {
     /// ```
     #[inline]
     pub async fn acquire_async_with<F: FnOnce()>(&self, begin_wait: F) {
-        self.acquire_async_with_internal(1, begin_wait).await;
-    }
-
-    /// Gets a permit from the semaphore synchronously.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use saa::Semaphore;
-    /// use std::sync::atomic::Ordering::Relaxed;
-    ///
-    /// let semaphore = Semaphore::default();
-    ///
-    /// semaphore.acquire_sync();
-    /// assert_eq!(semaphore.available_permits(Relaxed), Semaphore::MAX_PERMITS - 1);
-    /// ```
-    #[inline]
-    pub fn acquire_sync(&self) {
-        self.acquire_many_sync_with(1, || ());
+        self.acquire_many_async_with(1, begin_wait).await;
     }
 
     /// Gets multiple permits from the semaphore synchronously with a wait callback.
@@ -243,7 +271,27 @@ impl Semaphore {
     /// ```
     #[inline]
     pub async fn acquire_many_async(&self, count: usize) -> bool {
-        self.acquire_async_with_internal(count, || {}).await
+        self.acquire_many_async_with(count, || {}).await
+    }
+
+    /// Gets multiple permits from the semaphore synchronously.
+    ///
+    /// Returns `false` if the count exceeds [`Self::MAX_PERMITS`].
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use saa::Semaphore;
+    /// use std::sync::atomic::Ordering::Relaxed;
+    ///
+    /// let semaphore = Semaphore::default();
+    ///
+    /// assert!(semaphore.acquire_many_sync(11));
+    /// assert_eq!(semaphore.available_permits(Relaxed), Semaphore::MAX_PERMITS - 11);
+    /// ```
+    #[inline]
+    pub fn acquire_many_sync(&self, count: usize) -> bool {
+        self.acquire_many_sync_with(count, || ())
     }
 
     /// Gets multiple permits from the semaphore asynchronously with a wait callback.
@@ -268,27 +316,27 @@ impl Semaphore {
     /// ```
     #[inline]
     pub async fn acquire_many_async_with<F: FnOnce()>(&self, count: usize, begin_wait: F) -> bool {
-        self.acquire_async_with_internal(count, begin_wait).await
-    }
+        if count > Self::MAX_PERMITS {
+            return false;
+        }
+        let Ok(count) = u8::try_from(count) else {
+            return false;
+        };
+        loop {
+            let (result, state) = self.try_acquire_internal(count);
+            if result {
+                return true;
+            }
 
-    /// Gets multiple permits from the semaphore synchronously.
-    ///
-    /// Returns `false` if the count exceeds [`Self::MAX_PERMITS`].
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use saa::Semaphore;
-    /// use std::sync::atomic::Ordering::Relaxed;
-    ///
-    /// let semaphore = Semaphore::default();
-    ///
-    /// assert!(semaphore.acquire_many_sync(11));
-    /// assert_eq!(semaphore.available_permits(Relaxed), Semaphore::MAX_PERMITS - 11);
-    /// ```
-    #[inline]
-    pub fn acquire_many_sync(&self, count: usize) -> bool {
-        self.acquire_many_sync_with(count, || ())
+            let async_wait = WaitQueue::default();
+            let async_wait_pinned = async_wait.pin();
+            async_wait_pinned.construct(self, Opcode::Semaphore(count), false);
+            if self.try_push_wait_queue_entry(async_wait_pinned, state) {
+                begin_wait();
+                async_wait_pinned.await;
+                return true;
+            }
+        }
     }
 
     /// Gets multiple permits from the semaphore synchronously with a wait callback.
@@ -406,11 +454,7 @@ impl Semaphore {
                 pager.wait_queue().entry().set_result(0);
                 break;
             }
-
-            if self
-                .try_push_wait_queue_entry(pager.wait_queue(), state, || ())
-                .is_none()
-            {
+            if self.try_push_wait_queue_entry(pager.wait_queue(), state) {
                 break;
             }
         }
@@ -475,42 +519,6 @@ impl Semaphore {
         }
     }
 
-    /// Acquires permits asynchronously.
-    #[inline]
-    async fn acquire_async_with_internal<F: FnOnce()>(
-        &self,
-        count: usize,
-        mut begin_wait: F,
-    ) -> bool {
-        if count > Semaphore::MAX_PERMITS {
-            return false;
-        }
-        let Ok(count) = u8::try_from(count) else {
-            return false;
-        };
-        loop {
-            let (result, state) = self.try_acquire_internal(count);
-            if result {
-                return true;
-            }
-            debug_assert!(state & WaitQueue::ADDR_MASK != 0 || state & WaitQueue::DATA_MASK != 0);
-
-            let async_wait = pin!(WaitQueue::default());
-            async_wait
-                .as_ref()
-                .construct(self, Opcode::Semaphore(count), false);
-            if let Some(returned) =
-                self.try_push_wait_queue_entry(async_wait.as_ref(), state, begin_wait)
-            {
-                begin_wait = returned;
-                continue;
-            }
-
-            PinnedEntry(Pin::new(async_wait.entry())).await;
-            return true;
-        }
-    }
-
     /// Tries to acquire a permit.
     #[inline]
     fn try_acquire_internal(&self, count: u8) -> (bool, usize) {
