@@ -1,50 +1,74 @@
 use std::pin::pin;
+use std::sync::Arc;
 use std::sync::atomic::AtomicUsize;
-use std::sync::atomic::Ordering::Relaxed;
-use std::sync::{Arc, Barrier};
+use std::sync::atomic::Ordering::{Relaxed, Release};
 use std::thread;
 use std::time::Duration;
 
-use crate::{Gate, Lock, Pager, Semaphore, gate, lock};
+use crate::{Barrier, Gate, Lock, Pager, Semaphore, gate, lock};
 
 #[test]
 fn future_size() {
-    let limit = 196;
+    let limit = 168;
+    let limit_relaxed = 208;
     let lock = Lock::default();
 
-    let lock_fut = &lock.lock_async();
-    assert!(size_of_val(lock_fut) < limit);
+    let lock_fut_size = size_of_val(&lock.lock_async());
+    assert!(lock_fut_size <= limit, "{lock_fut_size}");
 
-    let lock_with_fut = &lock.lock_async_with(|| {});
-    assert!(size_of_val(lock_with_fut) < limit);
+    let lock_with_fut_size = size_of_val(&lock.lock_async_with(|| {}));
+    assert!(lock_with_fut_size <= limit, "{lock_with_fut_size}");
 
-    let share_fut = &lock.share_async();
-    assert!(size_of_val(share_fut) < limit);
+    let share_fut_size = size_of_val(&lock.share_async());
+    assert!(share_fut_size <= limit, "{share_fut_size}");
 
-    let share_with_fut = &lock.share_async_with(|| {});
-    assert!(size_of_val(share_with_fut) < limit);
+    let share_with_fut_size = size_of_val(&lock.share_async_with(|| {}));
+    assert!(share_with_fut_size <= limit, "{share_with_fut_size}");
+
+    let barrier = Barrier::default();
+
+    let barrier_fut_size = size_of_val(&barrier.wait_async());
+    assert!(barrier_fut_size <= limit_relaxed, "{barrier_fut_size}");
+
+    let barrier_with_fut_size = size_of_val(&barrier.wait_async_with(|| {}));
+    assert!(
+        barrier_with_fut_size <= limit_relaxed,
+        "{barrier_with_fut_size}"
+    );
 
     let semaphore = Semaphore::default();
 
-    let acquire_fut = &semaphore.acquire_async();
-    assert!(size_of_val(acquire_fut) < limit);
+    let acquire_fut_size = size_of_val(&semaphore.acquire_async());
+    assert!(acquire_fut_size <= limit_relaxed, "{acquire_fut_size}");
 
-    let acquire_with_fut = &semaphore.acquire_async_with(|| {});
-    assert!(size_of_val(acquire_with_fut) < limit);
+    let acquire_with_fut_size = size_of_val(&semaphore.acquire_async_with(|| {}));
+    assert!(
+        acquire_with_fut_size <= limit_relaxed,
+        "{acquire_with_fut_size}"
+    );
 
-    let acquire_many_fut = &semaphore.acquire_many_async(1);
-    assert!(size_of_val(acquire_many_fut) < limit);
+    let acquire_many_fut_size = size_of_val(&semaphore.acquire_many_async(1));
+    assert!(
+        acquire_many_fut_size <= limit_relaxed,
+        "{acquire_many_fut_size}"
+    );
 
-    let acquire_many_with_fut = &semaphore.acquire_many_async_with(1, || {});
-    assert!(size_of_val(acquire_many_with_fut) < limit);
+    let acquire_many_with_fut_size = size_of_val(&semaphore.acquire_many_async_with(1, || {}));
+    assert!(
+        acquire_many_with_fut_size <= limit_relaxed,
+        "{acquire_many_with_fut_size}"
+    );
 
     let gate = Gate::default();
 
-    let enter_fut = &gate.enter_async();
-    assert!(size_of_val(enter_fut) < limit);
+    let enter_fut_size = size_of_val(&gate.enter_async());
+    assert!(enter_fut_size <= limit_relaxed, "{enter_fut_size}");
 
-    let enter_with_fut = &gate.enter_async_with(|| {});
-    assert!(size_of_val(enter_with_fut) < limit);
+    let enter_with_fut_size = size_of_val(&gate.enter_async_with(|| {}));
+    assert!(
+        enter_with_fut_size <= limit_relaxed,
+        "{enter_with_fut_size}"
+    );
 }
 
 #[cfg_attr(miri, ignore = "Tokio is not compatible with Miri")]
@@ -180,7 +204,7 @@ fn lock_sync_wait_callback() {
     } else {
         Lock::MAX_SHARED_OWNERS
     };
-    let barrier = Arc::new(Barrier::new(num_threads + 1));
+    let barrier = Arc::new(Barrier::with_count(num_threads + 1));
     let lock = Arc::new(Lock::default());
 
     lock.lock_sync();
@@ -192,12 +216,12 @@ fn lock_sync_wait_callback() {
         threads.push(thread::spawn(move || {
             let result = if i % 7 == 3 {
                 lock.lock_sync_with(|| {
-                    barrier.wait();
+                    barrier.wait_sync();
                 });
                 lock.release_lock()
             } else {
                 lock.share_sync_with(|| {
-                    barrier.wait();
+                    barrier.wait_sync();
                 });
                 lock.release_share()
             };
@@ -205,7 +229,7 @@ fn lock_sync_wait_callback() {
         }));
     }
 
-    barrier.wait();
+    barrier.wait_sync();
     assert!(lock.release_lock());
 
     for thread in threads {
@@ -352,7 +376,7 @@ fn lock_poison_wait_sync() {
     } else {
         Lock::MAX_SHARED_OWNERS
     };
-    let barrier = Arc::new(Barrier::new(num_threads + 1));
+    let barrier = Arc::new(Barrier::with_count(num_threads + 1));
     let lock = Arc::new(Lock::default());
 
     lock.lock_sync();
@@ -364,17 +388,17 @@ fn lock_poison_wait_sync() {
         threads.push(thread::spawn(move || {
             if i % 2 == 0 {
                 assert!(!lock.lock_sync_with(|| {
-                    barrier.wait();
+                    barrier.wait_sync();
                 }));
             } else {
                 assert!(!lock.share_sync_with(|| {
-                    barrier.wait();
+                    barrier.wait_sync();
                 }));
             }
         }));
     }
 
-    barrier.wait();
+    barrier.wait_sync();
     assert!(lock.poison_lock());
 
     for thread in threads {
@@ -566,6 +590,30 @@ async fn lock_pager_wait_parallel() {
     }
 }
 
+#[test]
+fn barrier_sync() {
+    let num_threads = if cfg!(miri) { 4 } else { Barrier::MAX_TASKS };
+    let num_iters = if cfg!(miri) { 2 } else { 8 };
+    let barrier = Arc::new(Barrier::with_count(num_threads));
+
+    let mut threads = Vec::new();
+    for _ in 0..num_threads {
+        let barrier = barrier.clone();
+        threads.push(thread::spawn(move || {
+            for _ in 0..num_iters {
+                debug_assert!(barrier.count(Relaxed) <= num_threads);
+                barrier.wait_sync();
+            }
+            debug_assert!(barrier.count(Relaxed) <= num_threads);
+        }));
+    }
+
+    for thread in threads {
+        thread.join().unwrap();
+    }
+    assert_eq!(barrier.count(Relaxed), num_threads);
+}
+
 #[cfg_attr(miri, ignore = "Tokio is not compatible with Miri")]
 #[tokio::test(flavor = "multi_thread", worker_threads = 16)]
 async fn semaphore_async() {
@@ -656,7 +704,7 @@ fn semaphore_sync_wait_callback() {
     } else {
         Semaphore::MAX_PERMITS
     };
-    let barrier = Arc::new(Barrier::new(num_threads + 1));
+    let barrier = Arc::new(Barrier::with_count(num_threads));
     let semaphore = Arc::new(Semaphore::default());
 
     assert!(semaphore.acquire_many_sync(Semaphore::MAX_PERMITS));
@@ -668,12 +716,16 @@ fn semaphore_sync_wait_callback() {
         threads.push(thread::spawn(move || {
             let result = if i % 7 == 3 {
                 semaphore.acquire_sync_with(|| {
-                    barrier.wait();
+                    if i != 0 {
+                        barrier.wait_sync();
+                    }
                 });
                 semaphore.release()
             } else {
                 semaphore.acquire_many_sync_with(3, || {
-                    barrier.wait();
+                    if i != 0 {
+                        barrier.wait_sync();
+                    }
                 });
                 semaphore.release_many(3)
             };
@@ -681,7 +733,7 @@ fn semaphore_sync_wait_callback() {
         }));
     }
 
-    barrier.wait();
+    barrier.wait_sync();
     assert!(semaphore.release_many(Semaphore::MAX_PERMITS));
 
     for thread in threads {
@@ -1037,6 +1089,52 @@ async fn lock_chaos() {
     assert_eq!(check.load(Relaxed), 0);
 }
 
+#[cfg_attr(miri, ignore = "Tokio is not compatible with Miri")]
+#[tokio::test(flavor = "multi_thread", worker_threads = 16)]
+async fn barrier_chaos() {
+    let num_tasks = 16;
+    let num_iters = 2048;
+    let check = Arc::new(AtomicUsize::new(0));
+    let barrier = Arc::new(Barrier::with_count(num_tasks));
+
+    let mut threads = Vec::new();
+    let mut tasks = Vec::new();
+    for i in 0..num_tasks {
+        let check = check.clone();
+        let barrier = barrier.clone();
+        if i % 2 == 0 {
+            tasks.push(tokio::spawn(async move {
+                for i in 0..num_iters {
+                    debug_assert_eq!(check.load(Relaxed), i);
+                    if barrier.wait_async().await {
+                        check.fetch_add(1, Release);
+                    }
+                    barrier.wait_async().await;
+                }
+            }));
+        } else {
+            threads.push(thread::spawn(move || {
+                for i in 0..num_iters {
+                    debug_assert_eq!(check.load(Relaxed), i);
+                    if barrier.wait_sync() {
+                        check.swap(i + 1, Release);
+                    }
+                    barrier.wait_sync();
+                }
+            }));
+        }
+    }
+
+    for thread in threads {
+        thread.join().unwrap();
+    }
+    for task in tasks {
+        task.await.unwrap();
+    }
+
+    assert_eq!(check.load(Relaxed), num_iters);
+}
+
 #[cfg_attr(miri, ignore = "Tokio is not compatible with Miri")]
 #[tokio::test(flavor = "multi_thread", worker_threads = 16)]
 async fn semaphore_chaos() {
