@@ -59,7 +59,7 @@ pub(crate) struct Entry {
     state: AtomicU16,
     /// Indicates that the wait queue entry can be polled.
     ///
-    /// If the flag is set, the `drop` method will automatically release unknowlingly acquired
+    /// If the flag is set, the `drop` method will automatically release unknowingly acquired
     /// resources.
     pollable: std::sync::atomic::AtomicBool, // `Loom` is too slow when it is added to modeling.
     /// Monitors the result.
@@ -72,30 +72,13 @@ pub(crate) struct Entry {
     offset: u16,
 }
 
-/// Helper struct for pinning a [`WaitQueue`] to the stack and awaiting it without consuming it.
-pub(crate) struct PinnedEntry<'e>(pub(crate) Pin<&'e Entry>);
-
-/// Contextual data for asynchronous [`WaitQueue`].
-#[derive(Debug)]
-struct AsyncContext {
-    /// Waker to wake an executor when the result is ready.
-    waker: UnsafeCell<Option<Waker>>,
-}
-
-/// Contextual data for synchronous [`WaitQueue`].
-#[derive(Debug)]
-struct SyncContext {
-    /// Waker to wake an executor when the result is ready.
-    thread: UnsafeCell<Option<Thread>>,
-}
-
 /// Monitors the result.
 #[derive(Debug)]
 enum Monitor {
     /// Monitors asynchronously.
-    Async(AsyncContext),
+    Async(UnsafeCell<Option<Waker>>),
     /// Monitors synchronously.
-    Sync(SyncContext),
+    Sync(UnsafeCell<Option<Thread>>),
 }
 
 /// Static assertions.
@@ -142,13 +125,9 @@ impl WaitQueue {
         };
         let entry_ptr = Self::to_entry_ptr(anchor_ptr).cast_mut();
         let monitor = if is_sync {
-            Monitor::Sync(SyncContext {
-                thread: UnsafeCell::new(None),
-            })
+            Monitor::Sync(UnsafeCell::new(None))
         } else {
-            Monitor::Async(AsyncContext {
-                waker: UnsafeCell::new(None),
-            })
+            Monitor::Async(UnsafeCell::new(None))
         };
         unsafe {
             if had_entry {
@@ -188,9 +167,15 @@ impl WaitQueue {
         }
     }
 
-    /// Gets a pinned reference.
+    /// Gets a pinned reference from `self`.
     #[inline]
-    pub(crate) fn pinned_wait_queue<'l>(wait_queue_ptr: *const WaitQueue) -> Pin<&'l WaitQueue> {
+    pub(crate) const fn pin(&self) -> Pin<&WaitQueue> {
+        unsafe { Pin::new_unchecked(self) }
+    }
+
+    /// Gets a pinned reference from a pointer.
+    #[inline]
+    pub(crate) const fn pin_ptr<'l>(wait_queue_ptr: *const WaitQueue) -> Pin<&'l WaitQueue> {
         unsafe { Pin::new_unchecked(&*wait_queue_ptr) }
     }
 
@@ -287,6 +272,15 @@ impl Drop for WaitQueue {
     }
 }
 
+impl Future for Pin<&'_ WaitQueue> {
+    type Output = u8;
+
+    #[inline]
+    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
+        self.entry().poll_result_async(cx)
+    }
+}
+
 unsafe impl Send for WaitQueue {}
 unsafe impl Sync for WaitQueue {}
 
@@ -465,15 +459,15 @@ impl Entry {
             // A waker had been set before the result was set.
             unsafe {
                 match &self.monitor {
-                    Monitor::Async(async_context) => {
-                        if let Some(waker) = (*async_context.waker.get()).take() {
+                    Monitor::Async(waker) => {
+                        if let Some(waker) = (*waker.get()).take() {
                             self.state.fetch_or(Self::RESULT_FINALIZED, AcqRel);
                             waker.wake();
                             return;
                         }
                     }
-                    Monitor::Sync(sync_context) => {
-                        if let Some(thread) = (*sync_context.thread.get()).take() {
+                    Monitor::Sync(thread) => {
+                        if let Some(thread) = (*thread.get()).take() {
                             self.state.fetch_or(Self::RESULT_FINALIZED, AcqRel);
                             thread.unpark();
                             return;
@@ -485,56 +479,9 @@ impl Entry {
         self.state.fetch_or(Self::RESULT_FINALIZED, AcqRel);
     }
 
-    /// Polls the result, asynchronously.
-    pub(crate) fn poll_result_async(&self, cx: &mut Context<'_>) -> Poll<u8> {
-        let Monitor::Async(async_context) = &self.monitor else {
-            return Poll::Ready(Self::ERROR_WRONG_MODE);
-        };
-
-        if let Some(result) = self.try_consume_result() {
-            return Poll::Ready(result);
-        }
-
-        let mut this_waker = None;
-        let state = self.state.load(Acquire);
-        if state & Self::RESULT_SET == Self::RESULT_SET {
-            // No need to install the waker.
-            if let Some(result) = self.try_consume_result() {
-                return Poll::Ready(result);
-            }
-        } else if state & Self::WAKER_SET == Self::WAKER_SET {
-            // Replace the waker by clearing the flag first.
-            if self
-                .state
-                .compare_exchange_weak(state, state & !Self::WAKER_SET, AcqRel, Acquire)
-                .is_ok()
-            {
-                this_waker.replace(cx.waker().clone());
-            }
-        } else {
-            this_waker.replace(cx.waker().clone());
-        }
-
-        if let Some(waker) = this_waker {
-            unsafe {
-                (*async_context.waker.get()).replace(waker);
-            }
-            if self.state.fetch_or(Self::WAKER_SET, Release) & Self::RESULT_SET == Self::RESULT_SET
-            {
-                // The result has been set, so the waker will not be notified.
-                cx.waker().wake_by_ref();
-            }
-        } else {
-            // The waker is not set, so we need to wake the task.
-            cx.waker().wake_by_ref();
-        }
-
-        Poll::Pending
-    }
-
     /// Polls the result, synchronously.
     pub(crate) fn poll_result_sync(&self) -> u8 {
-        let Monitor::Sync(sync_context) = &self.monitor else {
+        let Monitor::Sync(thread) = &self.monitor else {
             return Self::ERROR_WRONG_MODE;
         };
 
@@ -563,9 +510,9 @@ impl Entry {
                 this_thread.replace(current());
             }
 
-            if let Some(thread) = this_thread {
+            if let Some(this_thread) = this_thread {
                 unsafe {
-                    (*sync_context.thread.get()).replace(thread);
+                    (*thread.get()).replace(this_thread);
                 }
                 if self.state.fetch_or(Self::WAKER_SET, Release) & Self::RESULT_SET
                     == Self::RESULT_SET
@@ -620,6 +567,53 @@ impl Entry {
         None
     }
 
+    /// Polls the result, asynchronously.
+    fn poll_result_async(&self, cx: &mut Context<'_>) -> Poll<u8> {
+        let Monitor::Async(waker) = &self.monitor else {
+            return Poll::Ready(Self::ERROR_WRONG_MODE);
+        };
+
+        if let Some(result) = self.try_consume_result() {
+            return Poll::Ready(result);
+        }
+
+        let mut this_waker = None;
+        let state = self.state.load(Acquire);
+        if state & Self::RESULT_SET == Self::RESULT_SET {
+            // No need to install the waker.
+            if let Some(result) = self.try_consume_result() {
+                return Poll::Ready(result);
+            }
+        } else if state & Self::WAKER_SET == Self::WAKER_SET {
+            // Replace the waker by clearing the flag first.
+            if self
+                .state
+                .compare_exchange_weak(state, state & !Self::WAKER_SET, AcqRel, Acquire)
+                .is_ok()
+            {
+                this_waker.replace(cx.waker().clone());
+            }
+        } else {
+            this_waker.replace(cx.waker().clone());
+        }
+
+        if let Some(this_waker) = this_waker {
+            unsafe {
+                (*waker.get()).replace(this_waker);
+            }
+            if self.state.fetch_or(Self::WAKER_SET, Release) & Self::RESULT_SET == Self::RESULT_SET
+            {
+                // The result has been set, so the waker will not be notified.
+                cx.waker().wake_by_ref();
+            }
+        } else {
+            // The waker is not set, so we need to wake the task.
+            cx.waker().wake_by_ref();
+        }
+
+        Poll::Pending
+    }
+
     /// Prepares for dropping `self`.
     ///
     /// The method cannot be implemented in `drop` because `Miri` treats the drop method in a
@@ -638,15 +632,5 @@ impl Entry {
     }
 }
 
-impl Future for PinnedEntry<'_> {
-    type Output = u8;
-
-    #[inline]
-    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
-        let this = self.get_mut();
-        this.0.poll_result_async(cx)
-    }
-}
-
 unsafe impl Send for Monitor {}
 unsafe impl Sync for Monitor {}
